
class Action:
    def __init__(self, time, effect):
        self.time = time
        self.effect = effect

class Bowl:
    def __init__(self, capacity):
        self.capacity = capacity
        self.amount = 0
        self.resource = 0
        self.mixed = False
        self.in_use = False

class StorageContext:
    def __init__(self, resource_id, switch_cost, max_amount, table, name):
        self.resource_id = resource_id
        self.switch_cost = switch_cost
        self.max_amount = max_amount
        self.table = table
        self.name = name

    def do(self, tick, speed):
        def effect():
            self.table[self.resource_id] += self.max_amount
        return Action(tick + speed, effect)

class BowlContext:
    def __init__(self, switch_cost, table, bowls, name):
        self.switch_cost = switch_cost
        self.bowls = bowls
        self.table = table
        self.name = name

    def do(self, bowl_id, resource_id, tick, speed):
        amount = 1
        assert 0 <= bowl_id < len(self.bowls), f"Bowl id {bowl_id} is not in range [{0}, {len(self.bowls)}["
        assert 1 <= resource_id < len(self.table), f"Resource id {bowl_id} is not in range [{1}, {len(self.table)}["
        assert self.table[0] >= amount, f"Trying to use {amount} of resource {0}, but only {self.table[0]} available"
        assert self.table[resource_id] >= amount, f"Trying to use {amount} of resource {resource_id}, but only {self.table[resource_id]} available"
        assert self.bowls[bowl_id].amount == 0 or self.bowls[bowl_id].resource == resource_id, f"Bowl {bowl_id} already has resource {self.bowls[bowl_id].resource}, can't add resource {resource_id} to it."
        assert self.bowls[bowl_id].amount == 0 or self.bowls[bowl_id].mixed == False, f"Bowl {bowl_id} is mixed can't add resources to it."
        assert self.bowls[bowl_id].in_use == False, f"Bowl {bowl_id} is already in use by a different action."
        assert self.bowls[bowl_id].amount + 2*amount <= self.bowls[bowl_id].capacity, f"Trying to add {2 * amount} resources to bowl {bowl_id}, but that exceeds the maximum capacity: {self.bowls[bowl_id].amount} + {2 * amount} > {self.bowls[bowl_id].capacity}"
        self.table[resource_id] -= amount;
        self.table[0] -= amount;
        self.bowls[bowl_id].mixed = False
        self.bowls[bowl_id].in_use = True
        self.bowls[bowl_id].resource = resource_id
        def finish(bowl_id=bowl_id):
            self.bowls[bowl_id].amount += 2 * amount
            self.bowls[bowl_id].in_use = False
        return Action(tick + speed * amount * 2, finish)

class MixerContext:
    def __init__(self, switch_cost, bowls, name):
        self.switch_cost = switch_cost
        self.bowls = bowls
        self.mixer_in_use = False
        self.name = name

    def do(self, bowl_id, tick, speed):
        assert 0 <= bowl_id < len(self.bowls), f"Bowl id {bowl_id} is not in range [{0}, {len(self.bowls)}["
        assert self.bowls[bowl_id].amount > 0 , f"Bowl {bowl_id} is empty."
        assert self.bowls[bowl_id].mixed == False, f"Bowl {bowl_id} is already mixed."
        assert self.bowls[bowl_id].in_use == False, f"Bowl {bowl_id} is already in use by a different action."
        assert self.mixer_in_use == False, "Mixer is already in use by a different player, bad luck."
        self.bowls[bowl_id].in_use = True
        self.mixer_in_use = True
        def finish(bowl_id=bowl_id):
            self.bowls[bowl_id].mixed = True
            self.bowls[bowl_id].in_use = False
            self.mixer_in_use = False
        return Action(tick + speed, finish)

class CupContext:
    def __init__(self, switch_cost, bowls, cups, max_cups, name):
        self.switch_cost = switch_cost
        self.bowls = bowls
        self.cups = cups
        self.max_cups = max_cups
        self.in_use = False
        self.name = name

    def do(self, bowl_id, tick, speed):
        amount = 1
        assert 0 <= bowl_id < len(self.bowls), f"Bowl id {bowl_id} is not in range [{0}, {len(self.bowls)}["
        assert sum(self.cups) + amount <= self.max_cups, f"Adding {amount} cups to the existing {sum(self.cups)} cups exceeds the maximum number of {self.max_cups} cups"
        assert self.bowls[bowl_id].amount >= amount, f"Bowl {bowl_id} only has {self.bowls[bowl_id].amount} / {amount} resources, which is not enough."
        assert self.bowls[bowl_id].mixed == True, f"Bowl {bowl_id} is not mixed."
        assert self.bowls[bowl_id].in_use == False, f"Bowl {bowl_id} is already in use by a different action."
        assert self.in_use == False, "There is already another mouse filling muffins into cups."
        self.bowls[bowl_id].in_use = True
        self.in_use = True
        def finish(bowl_id=bowl_id):
            self.cups[self.bowls[bowl_id].resource] += amount
            self.bowls[bowl_id].amount -= amount
            self.in_use = False
            self.bowls[bowl_id].in_use = False
        return Action(tick + amount * speed, finish)

class OvenContext:
    def __init__(self, switch_cost, baking_time, cups, muffin_handler, name):
        self.switch_cost = switch_cost
        self.baking_time = baking_time
        self.cups = cups
        self.baking_until = 0
        self.muffin_handler = muffin_handler
        self.in_use = False
        self.name = name

    def do(self, tick, speed):
        assert sum(self.cups) > 0, "There are no muffins to bake"
        assert self.in_use == False, "There is already another mouse filling the oven."
        assert tick >= self.baking_until , f"The oven is baking until tick {self.baking_until}, it is now {tick}."
        self.in_use = True
        self.muffin_handler.finished_muffins(self.cups, tick + speed + self.baking_time)
        self.cups[:] = [0] * len(self.cups)
        
        def finish(tick=tick):
            self.in_use = False
            self.baking_until = tick + self.baking_time
        return Action(tick + speed, finish)

