import argparse
import subprocess
import contexts
import json

class Mouse:
    def __init__(self, path, props, output):
        self.proc = subprocess.Popen(path, stdin=subprocess.PIPE, stdout=subprocess.PIPE, shell=True, universal_newlines=True)
        self.props = props # array with time to run each context action
        self.next_action_tick = 0
        self.context = -1
        self.pending_action = None
        self.output = output

    def __del__(self):
        self.proc.kill()
        
    def write(self, *line):
        try:
            print(*line, file=self.proc.stdin, flush=True)
        except:
            pass

    def read(self):
        self.output.append(self.proc.stdout.readline()[:-1])
        return self.output[-1]

    def execute_pending(self):
        if self.pending_action is not None:
            self.pending_action()
            self.pending_action = None
    
    def assign_action(self, action):
        self.pending_action = action.effect
        self.next_action_tick = action.time

class Game:
    def __init__(self, mice, contexts, order_list):
        self.contexts = contexts
        self.mice = mice
        for i, mouse in enumerate(mice):
            # send container details:
            # bowls:
            mouse.write(len(contexts[3].bowls), contexts[3].bowls[0].capacity)
            # cups:
            mouse.write(contexts[5].max_cups)
            # send context details:
            # dough: switch_cost, max_amount
            mouse.write(contexts[0].switch_cost, contexts[0].max_amount)
            # chocolate: switch_cost, max_amount
            mouse.write(contexts[1].switch_cost, contexts[1].max_amount)
            # lemons: switch_cost, max_amount
            mouse.write(contexts[2].switch_cost, contexts[2].max_amount)
            # bowl: switch_cost, bowl_capacity, num_bowls
            mouse.write(contexts[3].switch_cost)
            # mixer: switch_cost
            mouse.write(contexts[4].switch_cost)
            # cup: switch_cost
            mouse.write(contexts[5].switch_cost)
            # oven: switch_cost, baking_time
            mouse.write(contexts[6].switch_cost, contexts[6].baking_time)
            # send mice properties (time to execute each context action)
            mouse.write(len(mice))
            for m in mice:
                mouse.write(*m.props)
            # send own mouse id
            mouse.write(i)
            # send order list
            mouse.write(sum(len(x) for x in order_list))
            for resource, orders in enumerate(order_list):
                for time, amount in orders.items():
                    mouse.write(resource, time, amount)
            
            
    def update_containers(self, mouse):
        mouse.write(*self.contexts[0].table)
        for bowl in self.contexts[3].bowls:
            mouse.write(bowl.amount, bowl.resource, int(bowl.mixed and bowl.amount > 0), int(bowl.in_use))
        mouse.write(*self.contexts[5].cups[1:3])
        
    def run(self, end_tick):
        tick = 0
        while tick < end_tick:
            next_tick = end_tick
            for i, mouse in enumerate(self.mice):
                if mouse.next_action_tick <= tick:
                    mouse.execute_pending()
                    for m in self.mice:
                        m.write("T", tick, i)
                    self.update_containers(mouse)
                    action = list(map(int, mouse.read().split()))
                    assert action, f"No output from bot, it probably crashed"
                    assert 0 <= action[0] < 3, f"Action must be either 0 [switch context], 1 [execute current context] or 2 [idle], got {action[0]}"
                    if action[0] == 0:
                        assert 0 <= action[1] < len(self.contexts), f"The context id {action[1]} is not in the range [0, {len(self.contexts)}["
                        mouse.next_action_tick = tick + self.contexts[action[1]].switch_cost
                        def switch_context(target_context=action[1]):
                            mouse.context = target_context
                        mouse.assign_action(contexts.Action(time=mouse.next_action_tick, effect=switch_context))
                    if action[0] == 1:
                        assert mouse.context >= 0, "Mouse must be in a context before working in it"
                        mouse.assign_action(self.contexts[mouse.context].do(*action[1:], tick=tick, speed=mouse.props[mouse.context]))
                    if action[0] == 2:
                        mouse.next_action_tick = -1 if len(action) == 1 else action[1]
                        action = action[:1]
                    for m in self.mice:
                        m.write(*(["A"] + action + [mouse.next_action_tick]))
                    # update containers
                    for m in self.mice:
                        self.update_containers(m)
            tick = min([m.next_action_tick for m in self.mice if m.next_action_tick > 0] + [end_tick])
        for mouse in self.mice:
            mouse.write("E")

class MuffinCallback:
    def __init__(self, order_list):
        self.fulfilled = 0
        self.order_list = order_list # order_list[resource][time] = amount
        self.failed_muffins = 0
        self.completed_muffins = {}
        
    def finished_muffins(self, muffins, tick):
        for resource, muffin_cnt in enumerate(muffins):
            if resource >= len(self.order_list):
                continue
            while self.order_list[resource] and next(iter(self.order_list[resource])) < tick:
                self.failed_muffins += self.order_list[resource][next(iter(self.order_list[resource]))]
                self.order_list[resource].pop(next(iter(self.order_list[resource])))
            while self.order_list[resource] and muffin_cnt > 0:
                remove = min(muffin_cnt, self.order_list[resource][next(iter(self.order_list[resource]))])
                muffin_cnt -= remove
                if tick not in self.completed_muffins:
                    self.completed_muffins[tick] = 0
                self.completed_muffins[tick] += remove
                self.order_list[resource][next(iter(self.order_list[resource]))] -= remove
                if muffin_cnt > 0:
                    self.order_list[resource].pop(next(iter(self.order_list[resource])))

def run(exe_paths, game_cfg, mice_cfg, contexts_cfg):
    order_list = [{o["time"]: o["amount"] for o in game_cfg["orders"] if o["resource"] == r} for r in range(game_cfg["num_resources"])]
    muffin_callback = MuffinCallback(order_list)
    buffers = contexts_cfg["buffers_simple"]
    buffers.update({name: [globals()["contexts"].__dict__[cfg["type"]](**cfg["params"]) for i in range(cfg["count"])] for name, cfg in contexts_cfg["buffers"].items()})
    def make_context(cfg):
        params = dict(cfg["params"], **{name: buffers[buf] for name, buf in cfg["buffers"].items()})
        if "callback" in cfg:
            params[cfg["callback"]] = muffin_callback
        return globals()["contexts"].__dict__[cfg["context"]](**params)
    contexts = [make_context(cfg) for cfg in contexts_cfg["contexts"]]
    mice_output = [[] for m in mice_cfg if m["name"] in game_cfg["mice"]]
    mice = [Mouse(exe_paths[i], [mouse["skills"][c.name] for c in contexts], mice_output[i]) for i, mouse in enumerate(m for m in mice_cfg if m["name"] in game_cfg["mice"])]
    game = Game(mice, contexts, order_list)
    game.run(max(max(d.keys()) for d in order_list if d))
    return muffin_callback.completed_muffins, mice_output

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--exec', dest='exe', nargs='+', help='commands to run programs to evaluate, space separated')
    parser.add_argument('--config', dest='config', help='path to game.json')
    parser.add_argument('--contexts', dest='contexts', help='path to contexts.json')
    parser.add_argument('--mice', dest='mice', help='path to mice.json')
    parser.add_argument('--input', dest='input', help='path to input file downloaded from soi.ch')
    parser.add_argument('--output', dest='output', help='path to output file for uploading to soi.ch')

    args = parser.parse_args()
    assert any([args.config, args.contexts, args.mice]) !=  any([args.input, args.output]), "You can only use either --config, --contexts, --mice or --input, --output"
    assert args.exe, "Missing --exec argument"
    if args.config:
        game_cfg = json.load(open(args.config))
        contexts_cfg = json.load(open(args.contexts))
        mice_cfg = json.load(open(args.mice))
        muffins, _ = run(args.exe, game_cfg, mice_cfg, contexts_cfg)
    else:
        with open(args.input, 'r+') as f:
            with open(args.output, 'w') as of:
                for i, line in enumerate(f.readlines()[1:]):
                    cfg = json.loads(line)
                    muffins, mice_output = run(args.exe, cfg["game"], cfg["mice"], cfg["contexts"])
                    assert len(mice_output) == 1, "Only one mouse supported for this subtask"
                    of.write("Case #{}: {}\n".format(i, ";".join([str(sum(muffins.values()))] + mice_output[0])))
    if sum(muffins.values()) == 0:
        print("No muffins baked :(")
    elif sum(muffins.values()) == 1:
        print("That smells good, we baked a muffin.")
    else:
        print(f"Yay, we baked {sum(muffins.values())} muffins! :)")

if __name__ == '__main__':
    main()
